/* NickServ functions.
 *
 * Services is copyright (c) 1996-1998 Andy Church.
 *     E-mail: <achurch@dragonfire.net>
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "services.h"
#include "timeout.h"
#include "encrypt.h"


static NickInfo *nicklists[256];	/* One for each initial character */

const char s_NickServ[] = NICKSERV_NAME;

#include "ns-help.c"


/* For local timeout use: */
#define TO_COLLIDE	0	/* Collide the user with this nick */
#define TO_RELEASE	1	/* Release a collided nick */


static int is_on_access(User *u, NickInfo *ni);
static void alpha_insert_nick(NickInfo *ni);
static NickInfo *makenick(const char *nick);
static int delnick(NickInfo *ni);

static void collide(NickInfo *ni, int from_timeout);
static void release(NickInfo *ni, int from_timeout);

static void add_ns_timeout(NickInfo *ni, int type, time_t delay);
static void del_ns_timeout(NickInfo *ni, int type);

static void do_help(const char *source);
static void do_register(const char *source);
static void do_identify(const char *source);
static void do_drop(const char *source);
static void do_set(const char *source);
static void do_set_password(NickInfo *ni, char *param);
static void do_set_kill(NickInfo *ni, char *param);
static void do_set_secure(NickInfo *ni, char *param);
static void do_access(const char *source);
static void do_info(const char *source);
static void do_list(const char *source);
static void do_recover(const char *source);
static void do_release(const char *source);
static void do_ghost(const char *source);
static void do_status(const char *source);
static void do_getpass(const char *source);
static void do_forbid(const char *source);

/*************************************************************************/

#ifndef SKELETON

/* Display total number of registered nicks and info about each; or, if
 * a specific nick is given, display information about that nick (like
 * /msg NickServ INFO <nick>).  If count_only != 0, then only display the
 * number of registered nicks (the nick parameter is ignored).
 */

void listnicks(int count_only, const char *nick)
{
    long count = 0;
    NickInfo *ni;
    int i;

    if (count_only) {

	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next)
		++count;
	}
	printf("%d nicknames registered.\n", count);

    } else if (nick) {

	struct tm tm;
	char buf[512];

	if (!(ni = findnick(nick))) {
	    printf("%s not registered.\n");
	    return;
	} else if (ni->flags & NI_VERBOTEN) {
	    printf("%s is FORBIDden.\n");
	    return;
	}
	printf("%s is %s\n", nick, ni->last_realname);
	printf("Last seen address: %s\n", ni->last_usermask);
	tm = *localtime(&ni->time_registered);
	printf("  Time registered: %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	tm = *localtime(&ni->last_seen);
	printf("   Last seen time: %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	*buf = 0;
	if (ni->flags & NI_KILLPROTECT)
	    strcat(buf, "Kill protection");
	if (ni->flags & NI_SECURE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Security");
	}
	if (!*buf)
	    strcpy(buf, "None");
	printf("          Options: %s\n", buf);

    } else {

	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		printf("    %-20s  %s\n", ni->nick, ni->last_usermask);
		++count;
	    }
	}
	printf("%d nicknames registered.\n", count);

    }
}

#endif	/* !SKELETON */

/*************************************************************************/

/* Return information on memory use.  Assumes pointers are valid. */

void get_nickserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i, j;
    NickInfo *ni;
    char **accptr;

    for (i = 0; i < 256; i++) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    count++;
	    mem += sizeof(*ni);
	    if (ni->last_usermask)
		mem += strlen(ni->last_usermask)+1;
	    if (ni->last_realname)
		mem += strlen(ni->last_realname)+1;
	    mem += sizeof(char *) * ni->accesscount;
	    for (accptr=ni->access, j=0; j < ni->accesscount; accptr++, j++) {
		if (*accptr)
		    mem += strlen(*accptr)+1;
	    }
	}
    }
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/
/*************************************************************************/

/* Main NickServ routine. */

void nickserv(const char *source, char *buf)
{
    char *cmd, *s;

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;

    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_NickServ, source, "\1PING %s", s);

#ifdef SKELETON

    } else {
	notice(s_NickServ, source, "%s is currently offline.", s_NickServ);

    }

#else

    } else if (stricmp(cmd, "HELP") == 0) {
	do_help(source);

    } else if (stricmp(cmd, "REGISTER") == 0) {
	do_register(source);

    } else if (stricmp(cmd, "IDENTIFY") == 0) {
	do_identify(source);

    } else if (stricmp(cmd, "DROP") == 0) {
	do_drop(source);

    } else if (stricmp(cmd, "SET") == 0) {
	do_set(source);

    } else if (stricmp(cmd, "ACCESS") == 0) {
	do_access(source);

    } else if (stricmp(cmd, "INFO") == 0) {
	do_info(source);

    } else if (stricmp(cmd, "LIST") == 0) {
	do_list(source);

    } else if (stricmp(cmd, "RECOVER") == 0) {
	do_recover(source);

    } else if (stricmp(cmd, "RELEASE") == 0) {
	do_release(source);

    } else if (stricmp(cmd, "GHOST") == 0) {
	do_ghost(source);

    } else if (stricmp(cmd, "STATUS") == 0) {
	do_status(source);

    } else if (!is_services_admin(source)) {
	notice(s_NickServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_NickServ);

    } else if (stricmp(cmd, "GETPASS") == 0) {
	do_getpass(source);

    } else if (stricmp(cmd, "FORBID") == 0) {
	do_forbid(source);

    } else {
	notice(s_NickServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_NickServ);

    }

#endif	/* SKELETON */

}

/*************************************************************************/

#ifndef SKELETON


/* Load/save data files. */

void load_ns_dbase(void)
{
    FILE *f;
    int ver, i, j, len;
    NickInfo *ni;

    if (!(f = open_db(s_NickServ, NICKSERV_DB, "r")))
	return;
    switch (ver = get_file_version(f, NICKSERV_DB)) {
      case 3:
      case 2:
      case 1:
	for (i = 33; i < 256; ++i) {
	    while (fgetc(f) == 1) {
		ni = smalloc(sizeof(NickInfo));
		if (1 != fread(ni, sizeof(NickInfo), 1, f))
		    fatal_perror("Read error on %s", NICKSERV_DB);
#ifdef USE_ENCRYPTION
		if (!(ni->flags & NI_ENCRYPTEDPW)) {
		    if (encrypt_in_place(ni->pass, PASSMAX) < 0)
			fatal("%s: Can't encrypt `%s' nickname password!",
				s_NickServ, ni->nick);
		    ni->flags |= NI_ENCRYPTEDPW;
		}
#endif
		ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
		alpha_insert_nick(ni);
		ni->last_usermask = read_string(f, NICKSERV_DB);
		ni->last_realname = read_string(f, NICKSERV_DB);
		if (ni->accesscount) {
		    char **access;
		    access = smalloc(sizeof(char *) * ni->accesscount);
		    ni->access = access;
		    for (j = 0; j < ni->accesscount; ++j, ++access)
			*access = read_string(f, NICKSERV_DB);
		}
		ni->id_timestamp = 0;
		if (ver < 3) {
		    ni->memomax = DEF_MAX_MEMOS;
		    ni->flags |= NI_MEMO_SIGNON | NI_MEMO_RECEIVE;
		}
	    } /* while (fgetc(f) == 1) */
	} /* for (i) */
	break;
      default:
	fatal("Unsupported version number (%d) on %s", i, NICKSERV_DB);
    } /* switch (version) */
    fclose(f);
}

void save_ns_dbase(void)
{
    FILE *f;
    int i, j, len;
    NickInfo *ni;
    char **access;

    if (!(f = open_db(s_MemoServ, MEMOSERV_DB, "w")))
	return;
    for (i = 33; i < 256; ++i) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    fputc(1, f);
	    if (1 != fwrite(ni, sizeof(NickInfo), 1, f))
		fatal_perror("Write error on %s", NICKSERV_DB);
	    write_string(ni->last_usermask ? ni->last_usermask : "",
							f, NICKSERV_DB);
	    write_string(ni->last_realname ? ni->last_realname : "",
							f, NICKSERV_DB);
	    for (access = ni->access, j = 0; j < ni->accesscount; ++access, ++j)
		write_string(*access, f, NICKSERV_DB);
	}
	fputc(0, f);
    }
    fclose(f);
    remove(NICKSERV_DB ".save");
}

/*************************************************************************/

/* Check whether a user is on the access list of the nick they're using.
 * If not, send warnings as appropriate.  If so (and not NI_SECURE), update
 * last seen info.  Return 1 if the user is valid and recognized, 0
 * otherwise (note that this means an NI_SECURE nick will always return 0
 * from here). */

int validate_user(User *u)
{
    NickInfo *ni;
    int on_access;

    if (!(ni = findnick(u->nick)))
	return;

    if (ni->flags & NI_VERBOTEN) {
	notice(s_NickServ, u->nick,
		"This nickname may not be used.  Please choose another one.");
	notice(s_NickServ, u->nick,
		"If you do not change within one minute, you will be "
		"disconnected.");
	add_ns_timeout(ni, TO_COLLIDE, 60);
	return 0;
    }

    if (ni->id_timestamp != 0 && u->signon == ni->id_timestamp)
	;

    on_access = is_on_access(u, ni);

    if (!(ni->flags & NI_SECURE) && on_access) {
	ni->flags |= NI_RECOGNIZED;
	ni->last_seen = time(NULL);
	if (ni->last_usermask)
	    free(ni->last_usermask);
	ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	if (ni->last_realname)
	    free(ni->last_realname);
	ni->last_realname = sstrdup(u->realname);
	return 1;
    }

    if (ni->flags & NI_SECURE) {
	notice(s_NickServ, u->nick,
		"This nickname is registered and protected.  If it is your");
	notice(s_NickServ, u->nick,
		"nick, type \2/msg %s IDENTIFY \37password\37\2.  Otherwise,",
		s_NickServ);
	notice(s_NickServ, u->nick,
		"please choose a different nick.");
    } else {
	notice(s_NickServ, u->nick,
		"This nick is owned by someone else.  Please choose another.");
	notice(s_NickServ, u->nick,
		"(If this is your nick, type \2/msg %s IDENTIFY \37password\37\2.)",
		s_NickServ);
    }

    if ((ni->flags & NI_KILLPROTECT)
		&& !((ni->flags & NI_SECURE) && on_access)) {
	notice(s_NickServ, u->nick,
		"If you do not change within one minute, you will be "
		"disconnected.");
	add_ns_timeout(ni, TO_COLLIDE, 60);
    }

    return 0;
}

/*************************************************************************/

/* Cancel validation flags for a nick (i.e. when the user with that nick
 * signs off or changes nicks).  Also cancels any impending collide. */

void cancel_user(User *u)
{
    NickInfo *ni;

    if (ni = findnick(u->nick)) {
	ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	del_ns_timeout(ni, TO_COLLIDE);
    }
}

/*************************************************************************/

/* Remove all nicks which have expired. */

void expire_nicks()
{
    NickInfo *ni, *next;
    int i;
    const time_t expire_time = NICK_EXPIRE*24*60*60;
    time_t now = time(NULL);

    for (i = 33; i < 256; ++i) {
	for (ni = nicklists[i]; ni; ni = next) {
	    next = ni->next;
	    if (now - ni->last_seen >= expire_time
					&& !(ni->flags & NI_VERBOTEN)) {
		log("Expiring nickname %s", ni->nick);
		delnick(ni);
	    }
	}
    }
}

/*************************************************************************/

/* Return the NickInfo structure for the given nick, or NULL if the nick
 * isn't registered. */

NickInfo *findnick(const char *nick)
{
    NickInfo *ni;

    for (ni = nicklists[tolower(*nick)]; ni; ni = ni->next) {
	if (stricmp(ni->nick, nick) == 0)
	    return ni;
    }
    return NULL;
}

/*************************************************************************/
/*********************** NickServ private routines ***********************/
/*************************************************************************/

/* Is the given user's address on the given nick's access list?  Return 1
 * if so, 0 if not. */

static int is_on_access(User *u, NickInfo *ni)
{
    int i;
    char *buf;

    i = strlen(u->username);
    buf = smalloc(i + strlen(u->host) + 2);
    sprintf(buf, "%s@%s", u->username, u->host);
    strlower(buf+i+1);
    for (i = 0; i < ni->accesscount; ++i) {
	if (match_wild(ni->access[i], buf)) {
	    free(buf);
	    return 1;
	}
    }
    free(buf);
    return 0;
}

/*************************************************************************/

/* Insert a nick alphabetically into the database. */

static void alpha_insert_nick(NickInfo *ni)
{
    NickInfo *ni2, *ni3;
    char *nick = ni->nick;

    for (ni3 = NULL, ni2 = nicklists[tolower(*nick)];
			ni2 && stricmp(ni2->nick, nick) < 0;
			ni3 = ni2, ni2 = ni2->next)
	;
    ni->prev = ni3;
    ni->next = ni2;
    if (!ni3)
	nicklists[tolower(*nick)] = ni;
    else
	ni3->next = ni;
    if (ni2)
	ni2->prev = ni;
}

/*************************************************************************/

/* Add a nick to the database.  Returns a pointer to the new NickInfo
 * structure if the nick was successfully registered, NULL otherwise.
 * Assumes nick does not already exist. */

static NickInfo *makenick(const char *nick)
{
    NickInfo *ni;
    User *u;

    ni = scalloc(sizeof(NickInfo), 1);
    strscpy(ni->nick, nick, NICKMAX);
    alpha_insert_nick(ni);
    return ni;
}

/*************************************************************************/

/* Remove a nick from the NickServ database.  Return 1 on success, 0
 * otherwise.  Also deletes the nick from any channel access lists it is
 * on. */

static int delnick(NickInfo *ni)
{
    int i;

    if (ni->next)
	ni->next->prev = ni->prev;
    if (ni->prev)
	ni->prev->next = ni->next;
    else
	nicklists[tolower(*ni->nick)] = ni->next;
    if (ni->last_usermask)
	free(ni->last_usermask);
    if (ni->last_realname)
	free(ni->last_realname);
    if (ni->access) {
	for (i = 0; i < ni->accesscount; ++i)
	    free(ni->access[i]);
	free(ni->access);
    }
    cs_remove_nick(ni->nick);
    clear_memos(ni->nick);
    free(ni);
    return 1;
}

/*************************************************************************/

/* Collide a nick. */

static void collide(NickInfo *ni, int from_timeout)
{
    if (!from_timeout)
	del_ns_timeout(ni, TO_COLLIDE);
    kill_user(s_NickServ, ni->nick, "Nick kill enforced");
    send_cmd(NULL, "NICK %s %d 1 enforcer %s %s :%s Enforcement",
		ni->nick, time(NULL), services_host, server_name, s_NickServ);
    ni->flags |= NI_KILL_HELD;
    add_ns_timeout(ni, TO_RELEASE, RELEASE_TIMEOUT);
}

/*************************************************************************/

/* Release hold on a nick. */

static void release(NickInfo *ni, int from_timeout)
{
    if (!from_timeout)
	del_ns_timeout(ni, TO_RELEASE);
    send_cmd(ni->nick, "QUIT");
    ni->flags &= ~NI_KILL_HELD;
}

/*************************************************************************/
/*************************************************************************/

static struct my_timeout {
    struct my_timeout *next, *prev;
    NickInfo *ni;
    Timeout *to;
    int type;
} *my_timeouts;

/*************************************************************************/

/* Remove a collide/release timeout from our private list. */

static void rem_ns_timeout(NickInfo *ni, int type)
{
    struct my_timeout *t, *t2;

    t = my_timeouts;
    while (t) {
	if (t->ni == ni && t->type == type) {
	    t2 = t->next;
	    if (t->next)
		t->next->prev = t->prev;
	    if (t->prev)
		t->prev->next = t->next;
	    else
		my_timeouts = t->next;
	    free(t);
	    t = t2;
	} else {
	    t = t->next;
	}
    }
}

/*************************************************************************/

/* Collide a nick on timeout. */

static void timeout_collide(Timeout *t)
{
    NickInfo *ni = t->data;
    User *u;

    rem_ns_timeout(ni, TO_COLLIDE);
    /* If they identified or don't exist anymore, don't kill them. */
    if (ni->flags & NI_IDENTIFIED
		|| !(u = finduser(ni->nick))
		|| u->my_signon > t->settime)
	return;
    /* The RELEASE timeout will always add to the beginning of the
     * list, so we won't see it.  Which is fine because it can't be
     * triggered yet anyway. */
    collide(ni, 1);
}

/*************************************************************************/

/* Release a nick on timeout. */

static void timeout_release(Timeout *t)
{
    NickInfo *ni = t->data;

    rem_ns_timeout(ni, TO_RELEASE);
    release(ni, 1);
}

/*************************************************************************/

/* Add a collide/release timeout. */

static void add_ns_timeout(NickInfo *ni, int type, time_t delay)
{
    Timeout *to;
    struct my_timeout *t;
    void (*timeout_routine)(Timeout *);

    if (type == TO_COLLIDE)
	timeout_routine = timeout_collide;
    else if (type == TO_RELEASE)
	timeout_routine = timeout_release;
    else {
	log("NickServ: unknown timeout type %d!  ni=%p (%s), delay=%d",
		type, ni, ni->nick, delay);
	return;
    }
    to = add_timeout(delay, timeout_routine);
    to->data = ni;
    t = smalloc(sizeof(*t));
    t->next = my_timeouts;
    my_timeouts = t;
    t->prev = NULL;
    t->ni = ni;
    t->to = to;
    t->type = type;
}

/*************************************************************************/

/* Delete a collide/release timeout. */

static void del_ns_timeout(NickInfo *ni, int type)
{
    struct my_timeout *t, *t2;

    t = my_timeouts;
    while (t) {
	if (t->ni == ni && t->type == type) {
	    t2 = t->next;
	    if (t->next)
		t->next->prev = t->prev;
	    if (t->prev)
		t->prev->next = t->next;
	    else
		my_timeouts = t->next;
	    del_timeout(t->to);
	    free(t);
	    t = t2;
	} else {
	    t = t->next;
	}
    }
}

/*************************************************************************/
/*********************** NickServ command routines ***********************/
/*************************************************************************/

/* Return a help message. */

static void do_help(const char *source)
{
    char *cmd = strtok(NULL, "");
    char buf[BUFSIZE];
    int is_servadmin = is_services_admin(source);

    if (cmd && is_servadmin) {

	if (stricmp(cmd, "DROP") == 0) {
	    notice_list(s_NickServ, source, oper_drop_help);
	    return;
	} else if (stricmp(cmd, "GETPASS") == 0) {
	    notice_list(s_NickServ, source, getpass_help);
	    return;
	} else if (stricmp(cmd, "FORBID") == 0) {
	    notice_list(s_NickServ, source, forbid_help);
	    return;
	}

    }

    snprintf(buf, sizeof(buf), "%s%s", s_NickServ, cmd ? " " : "");
    strscpy(buf+strlen(buf), cmd ? cmd : "", sizeof(buf)-strlen(buf));
    helpserv(s_NickServ, source, buf);
    if (!cmd && is_servadmin)
	notice_list(s_NickServ, source, ns_servadmin_help);
}

/*************************************************************************/

/* Register a nick. */

static void do_register(const char *source)
{
    NickInfo *ni;
    User *u;
    char *pass = strtok(NULL, " ");

    if (readonly) {
	notice(s_NickServ, source,
		"Sorry, nickname registration is temporarily disabled.");
	return;
    }

    if (!pass) {

	notice(s_NickServ, source, "Syntax: \2REGISTER \37password\37\2");
	notice(s_NickServ, source,
			"\2/msg %s HELP REGISTER\2 for more information.",
			s_NickServ);

    } else if (!(u = finduser(source))) {

	log("%s: Can't register nick %s: nick not online", s_NickServ, source);
	notice(s_NickServ, source, "Sorry, registration failed.");

    } else if (ni = findnick(source)) {

	if (ni->flags & NI_VERBOTEN) {
	    log("%s: %s@%s tried to register FORBIDden nick %s", s_NickServ,
			u->username, u->host, source);
	    notice(s_NickServ, source,
			"Nickname \2%s\2 may not be registered.", source);
	} else {
	    notice(s_NickServ, source,
			"Nickname \2%s\2 is already registered!", source);
	}

    } else if (stricmp(source, pass) == 0 || strlen(pass) < 5) {

	notice(s_NickServ, source,
		"Please try again with a more obscure password.  Passwords "
		"should be at least five characters long, should not be "
		"something easily guessed (e.g. your real name or your "
		"nick), and cannot contain the space or tab characters.  "
		"\2/msg %s HELP REGISTER\2 for more information.", s_NickServ);

    } else {

	if (ni = makenick(source)) {
#ifdef USE_ENCRYPTION
	    int len = strlen(pass);
	    if (len > PASSMAX)
		len = PASSMAX;
	    if (encrypt(pass, len+1, ni->pass, PASSMAX) < 0) {
		memset(pass, 0, strlen(pass));
		log("%s: Failed to encrypt password for %s (register)",
			s_NickServ, source);
		notice(s_NickServ, source, "Sorry, registration failed.");
		return;
	    }
	    memset(pass, 0, strlen(pass));
	    ni->flags = NI_ENCRYPTEDPW | NI_IDENTIFIED | NI_RECOGNIZED;
#else
	    strscpy(ni->pass, pass, PASSMAX);
	    ni->flags = NI_IDENTIFIED | NI_RECOGNIZED;
#endif
	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	    ni->last_realname = sstrdup(u->realname);
	    ni->time_registered = ni->last_seen = time(NULL);
	    ni->accesscount = 1;
	    ni->access = smalloc(sizeof(char *));
	    ni->access[0] = create_mask(u);
	    /* Pre-3.0.1: Memos weren't cleared on a dropped nick.  Do
	     * this to make sure nobody gets someone else's memos. */
	    clear_memos(ni->nick);
	    log("%s: `%s' registered by %s@%s", s_NickServ,
		    source, u->username, u->host);
	    notice(s_NickServ, source,
		    "Nickname %s registered under your account: %s",
		    source, ni->access[0]);
#ifndef USE_ENCRYPTION
	    notice(s_NickServ, source,
		    "Your password is \2%s\2 - remember this for later use.",
		    pass);
#endif
	} else {
	    notice(s_NickServ, source,
		"Sorry, couldn't register your nickname.");
	}

    }

}

/*************************************************************************/

static void do_identify(const char *source)
{
    char *pass = strtok(NULL, " ");
    NickInfo *ni = findnick(source);
    User *u;
    int res;

    if (!pass) {

	notice(s_NickServ, source, "Syntax: \2IDENTIFY \37password\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.",
		s_NickServ);

    } else if (!ni) {

	notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!(u = finduser(source))) {

	log("%s: IDENTIFY from nonexistent nick %s", s_NickServ, source);
	notice(s_NickServ, source, "Sorry, identification failed.");

    } else if (!(res = check_password(pass, ni->pass))) {

	log("%s: Failed IDENTIFY for %s!%s@%s",
		s_NickServ, source, u->username, u->host);
	notice(s_NickServ, source, "Password incorrect.");
	bad_password(u);

    } else if (res == -1) {

	notice(s_NickServ, source, "Sorry, identification failed.");

    } else {

	ni->flags |= NI_IDENTIFIED;
	ni->id_timestamp = u->signon;
	if (!(ni->flags & NI_RECOGNIZED)) {
	    ni->last_seen = time(NULL);
	    if (ni->last_usermask)
		free(ni->last_usermask);
	    ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
	    if (ni->last_realname)
		free(ni->last_realname);
	    ni->last_realname = sstrdup(u->realname);
	    log("%s: %s!%s@%s identified for nick %s", s_NickServ,
			source, u->username, u->host, source);
	}
	notice(s_NickServ, source,
		"Password accepted - you are now recognized.");
	if (!(ni->flags & NI_RECOGNIZED))
	    check_memos(source);

    }
}

/*************************************************************************/

static void do_drop(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);;

    if (readonly && !is_services_admin(source)) {
	notice(s_NickServ, source,
		"Sorry, nickname de-registration is temporarily disabled.");
	return;
    }

    if (!is_services_admin(source) && nick) {

	notice(s_NickServ, source, "Syntax: \2DROP\2");
	notice(s_NickServ, source,
		"\2/msg %s DROP\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick ? nick : source))) {

	if (nick)
	    notice(s_NickServ, source, "Nick %s isn't registered.", nick);
	else
	    notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!nick && (!u || !(ni->flags & NI_IDENTIFIED))) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: DROP from nonexistent user %s", s_NickServ, source);

    } else if (nick && !u) {

	log("%s: DROP %s from nonexistent oper %s", s_NickServ, nick, source);
	notice(s_NickServ, source, "Can't find your user record!");

    } else {

	if (readonly) {
	    notice(s_NickServ, source,
		"Warning: Services is in read-only mode.  Changes will not be saved.");
	}
	delnick(ni);
	log("%s: %s!%s@%s dropped nickname %s", s_NickServ,
		source, u->username, u->host, nick ? nick : source);
	if (nick)
	    notice(s_NickServ, source, "Nickname %s has been dropped.", nick);
	else
	    notice(s_NickServ, source, "Your nickname has been dropped.");

    }
}

/*************************************************************************/

static void do_set(const char *source)
{
    char *cmd    = strtok(NULL, " ");
    char *param  = strtok(NULL, " ");
    NickInfo *ni = findnick(source);
    User *u;

    if (readonly) {
	notice(s_NickServ, source,
		"Sorry, nickname option setting is temporarily disabled.");
	return;
    }

    if (!param) {

	notice(s_NickServ, source,
		"Syntax: \2SET \37option\37 \37parameters\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(source))) {

	notice(s_NickServ, source, "Your nickname is not registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "PASSWORD") == 0) {

	do_set_password(ni, param);

    } else if (stricmp(cmd, "KILL") == 0) {

	do_set_kill(ni, param);

    } else if (stricmp(cmd, "SECURE") == 0) {

	do_set_secure(ni, param);

    } else {

	notice(s_NickServ, source,
			"Unknown SET option \2%s\2.", strupper(cmd));

    }
}

/*************************************************************************/

static void do_set_password(NickInfo *ni, char *param)
{
#ifdef USE_ENCRYPTION
    int len = strlen(param);
    if (len > PASSMAX)
	len = PASSMAX;
    if (encrypt(param, len, ni->pass, PASSMAX) < 0) {
	memset(param, 0, strlen(param));
	log("%s: Failed to encrypt password for %s (set)",
		s_NickServ, ni->nick);
	notice(s_NickServ, ni->nick, "Sorry, couldn't change password.");
	return;
    }
    memset(param, 0, strlen(param));
    notice(s_NickServ, ni->nick, "Password changed.");
#else
    strscpy(ni->pass, param, PASSMAX);
    notice(s_NickServ, ni->nick, "Password changed to \2%s\2.", param);
#endif
}

/*************************************************************************/

static void do_set_kill(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	ni->flags |= NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection is now \2ON\2.");

    } else if (stricmp(param, "OFF") == 0) {

	ni->flags &= ~NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection is now \2OFF\2.");

    } else {

	notice(s_NickServ, source, "Syntax: \2SET KILL {ON|OFF}\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET KILL\2 for more information.", s_NickServ);
    }
}

/*************************************************************************/

static void do_set_secure(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	ni->flags |= NI_SECURE;
	notice(s_NickServ, source, "Secure option is now \2ON\2.");

    } else if (stricmp(param, "OFF") == 0) {

	ni->flags &= ~NI_SECURE;
	notice(s_NickServ, source, "Secure option is now \2OFF\2.");

    } else {

	notice(s_NickServ, source, "Syntax: \2SET SECURE {ON|OFF}\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET SECURE\2 for more information.",
		s_NickServ);
    }
}

/*************************************************************************/

static void do_access(const char *source)
{
    char *cmd = strtok(NULL, " ");
    char *mask = strtok(NULL, " ");
    NickInfo *ni;
    User *u;
    int i;
    char **access;

    if (cmd && stricmp(cmd, "LIST") == 0 && mask && is_services_admin(source)) {

	if (!(ni = findnick(mask))) {
	    notice(s_NickServ, source, "Nick \2%s\2 not registered!", mask);
	    return;
	}
	notice(s_NickServ, source, "Access list for \2%s\2:", mask);
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i)
	    notice(s_NickServ, source, "    %s", *access);

    } else if (!cmd || ((stricmp(cmd,"LIST")==0) ? !!mask : !mask)) {

	notice(s_NickServ, source,
		"Syntax: \2ACCESS {ADD|DEL|LIST} [\37mask\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    } else if (mask && !strchr(mask, '@')) {

	notice(s_NickServ, source,
		"Mask must be in the form \37user\37@\37host\37.");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(source))) {

	notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "ADD") == 0) {

	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (strcmp(*access, mask) == 0) {
		notice(s_NickServ, source,
			"Mask \2%s\2 already present on your access list.",
			*access);
		return;
	    }
	}
	++ni->accesscount;
	ni->access = srealloc(ni->access, sizeof(char *) * ni->accesscount);
	ni->access[ni->accesscount-1] = sstrdup(mask);
	notice(s_NickServ, source, "\2%s\2 added to your access list.", mask);

    } else if (stricmp(cmd, "DEL") == 0) {

	/* First try for an exact match; then, a case-insensitive one. */
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (strcmp(*access, mask) == 0)
		break;
	}
	if (i == ni->accesscount) {
	    for (access = ni->access, i = 0; i < ni->accesscount;
							++access, ++i) {
		if (stricmp(*access, mask) == 0)
		    break;
	    }
	}
	if (i == ni->accesscount) {
	    notice(s_NickServ, source,
			"\2%s\2 not found on your access list.", mask);
	    return;
	}
	notice(s_NickServ, source,
		"\2%s\2 deleted from your access list.", *access);
	free(*access);
	--ni->accesscount;
	if (i < ni->accesscount)	/* if it wasn't the last entry... */
	    bcopy(access+1, access, (ni->accesscount-i) * sizeof(char *));
	if (ni->accesscount)		/* if there are any entries left... */
	    ni->access = srealloc(ni->access, ni->accesscount * sizeof(char *));
	else {
	    free(ni->access);
	    ni->access = NULL;
	}

    } else if (stricmp(cmd, "LIST") == 0) {

	notice(s_NickServ, source, "Access list:");
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (mask && !match_wild(mask, *access))
		continue;
	    notice(s_NickServ, source, "    %s", *access);
	}

    } else {

	notice(s_NickServ, source,
		"Syntax: \2ACCESS {ADD|DEL|LIST} [\37mask\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    }
}

/*************************************************************************/

static void do_info(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;

    if (!nick) {
	notice(s_NickServ, source, "Syntax: \2INFO \37nick\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick \2%s\2 isn't registered.", nick);

    } else if (ni->flags & NI_VERBOTEN) {
	notice(s_NickServ, source,
		"Nick \2%s\2 may not be registered or used.", nick);

    } else {
	struct tm tm;
	char buf[512];

	notice(s_NickServ, source,
		"%s is %s\n", nick, ni->last_realname);
	notice(s_NickServ, source,
		"Last seen address: %s\n", ni->last_usermask);
	tm = *localtime(&ni->time_registered);
	notice(s_NickServ, source,
		"  Time registered: %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	tm = *localtime(&ni->last_seen);
	notice(s_NickServ, source,
		"   Last seen time: %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	*buf = 0;
	if (ni->flags & NI_KILLPROTECT)
	    strcat(buf, "Kill protection");
	if (ni->flags & NI_SECURE) {
	    if (*buf)
		strcat(buf, ", ");
	    strcat(buf, "Security");
	}
	if (!*buf)
	    strcpy(buf, "None");
	notice(s_NickServ, source, "          Options: %s", buf);

    }
}

/*************************************************************************/

static void do_list(const char *source)
{
    char *pattern = strtok(NULL, " ");
    NickInfo *ni;
    int nnicks, i;
    char buf[BUFSIZE];

    if (!pattern) {

	notice(s_NickServ, source, "Syntax: \2LIST \37pattern\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP LIST\2 for more information.", s_NickServ);

    } else {

	nnicks = 0;
	notice(s_NickServ, source, "List of entries matching \2%s\2:",
		pattern);
	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		if (ni->flags & NI_VERBOTEN)
		    continue;
		snprintf(buf, sizeof(buf), "%-20s  %s", ni->nick, ni->last_usermask);
		if (stricmp(pattern, ni->nick) == 0 ||
					match_wild_nocase(pattern, buf)) {
		    if (++nnicks <= 50)
			notice(s_NickServ, source, "    %s", buf);
		}
	    }
	}
	notice(s_NickServ, source, "End of list - %d/%d matches shown.",
					nnicks>50 ? 50 : nnicks, nnicks);
    }

}

/*************************************************************************/

static void do_recover(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
		"Syntax: \2RECOVER \37nickname\37 [\37password\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP RECOVER\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!finduser(nick)) {

	notice(s_NickServ, source, "Nick %s isn't currently in use.");

    } else if (stricmp(nick, source) == 0) {

	notice(s_NickServ, source, "You can't recover yourself.");

    } else if (u &&
	    (pass ? check_password(pass,ni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	collide(ni, 0);
	notice(s_NickServ, source, "User claiming your nick has been killed.");
	notice(s_NickServ, source,
		"\2/msg %s RELEASE\2 to get it back before the one-minute "
		"timeout.", s_NickServ);

    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: RECOVER: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: RECOVER: invalid password for %s by %s!%s@%s",
			s_NickServ, source, u->username, u->host);
	    bad_password(u);
	}

    }
}

/*************************************************************************/

static void do_release(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
		"Syntax: \2RELEASE \37nickname\37 [\37password\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP RELEASE\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!(ni->flags & NI_KILL_HELD)) {

	notice(s_NickServ, source, "Nick %s isn't being held.", nick);

    } else if (u &&
	    (pass ? check_password(pass,ni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	release(ni, 0);
	notice(s_NickServ, source,
		"Services' hold on your nick has been released.");

    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: RELEASE: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: RELEASE: invalid password for %s by %s!%s@%s",
			s_NickServ, source, u->username, u->host);
	    bad_password(u);
	}

    }
}

/*************************************************************************/

static void do_ghost(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
		"Syntax: \2GHOST \37nickname\37 [\37password\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP GHOST\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!finduser(nick)) {

	notice(s_NickServ, source, "Nick %s isn't currently in use.", nick);

    } else if (stricmp(nick, source) == 0) {

	notice(s_NickServ, source, "You can't ghost yourself!");

    } else if (u &&
	    (pass ? check_password(pass,ni->pass)==1 :
	     !(ni->flags & NI_SECURE) && is_on_access(u,ni))) {

	char buf[NICKMAX+32];

	snprintf(buf, sizeof(buf), "GHOST command used by %s", source);
	kill_user(s_NickServ, nick, buf);
	notice(s_NickServ, source, "Ghost with your nick has been killed.");

    } else {

	notice(s_NickServ, source, "Access denied.");
	if (!u)
	    log("%s: GHOST: source user %s not found!", s_NickServ, source);
	else if (pass) {
	    log("%s: GHOST: invalid password for %s by %s!%s@%s",
			s_NickServ, source, u->username, u->host);
	    bad_password(u);
	}

    }
}

/*************************************************************************/

static void do_status(const char *source)
{
    char *nick;
    NickInfo *ni;
    int i = 0;

    while ((nick = strtok(NULL, " ")) && (i++ < 16)) {
	if (!(ni = findnick(nick)) || !finduser(nick))
	    notice(s_NickServ, source, "%s 0", nick);
	else if (ni->flags & NI_IDENTIFIED)
	    notice(s_NickServ, source, "%s 3", nick);
	else if (ni->flags & NI_RECOGNIZED)
	    notice(s_NickServ, source, "%s 2", nick);
	else
	    notice(s_NickServ, source, "%s 1", nick);
    }
}

/*************************************************************************/

static void do_getpass(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

#ifdef USE_ENCRYPTION
    notice(s_NickServ, source,
		"GETPASS command unavailable when encryption is in use.");
#else
    if (!nick) {
	notice(s_NickServ, source, "Syntax: \2GETPASS \37nickname\37\2");
    } else if (!u) {
	notice(s_NickServ, source, "Couldn't get your user info!");
    } else if (!(ni = findnick(nick))) {
	notice(s_NickServ, source, "Nick %s isn't registered.", nick);
    } else {
	log("%s: %s!%s@%s used GETPASS on %s",
		s_NickServ, source, u->username, u->host, nick);
	wallops(s_NickServ, "\2%s\2 used GETPASS on \2%s\2", source, nick);
	notice(s_NickServ, source, "Password for %s is \2%s\2.",
		nick, ni->pass);
    }
#endif
}

/*************************************************************************/

static void do_forbid(const char *source)
{
    NickInfo *ni;
    char *nick = strtok(NULL, " ");

    if (!nick) {
	notice(s_NickServ, source, "Syntax: \2FORBID \37nickname\37\2");
	return;
    }
    if (readonly) {
	notice(s_NickServ, source,
	    "Warning: Services is in read-only mode; changes will not be saved.");
    }
    if (ni = findnick(nick))
	delnick(ni);
    if (ni = makenick(nick)) {
	ni->flags |= NI_VERBOTEN;
	log("%s: %s set FORBID for nick %s", s_NickServ, source, nick);
	notice(s_NickServ, source, "Nick \2%s\2 is now FORBIDden.", nick);
    } else {
	log("%s: Valid FORBID for %s by %s failed", s_NickServ,
		nick, source);
	notice(s_NickServ, source, "Couldn't FORBID nick \2%s\2!", nick);
    }
}

/*************************************************************************/

#endif	/* !SKELETON */
